/*  abcat.c  --  adaptive, buffering cat

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Bishop Brock
IBM Austin Research Laboratory
11501 Burnet Rd.  MS/9460
Austin, TX 78757
(512) 838-0149
brock@austin.ibm.com

The abcat program provides a buffered stream copy function whose
buffering behavior adapts to the incoming input stream. This program
was designed to intelligently buffer large amounts of output sent by
an interactive program (ACL2) to be displayed in Emacs.  By buffering
fast-arriving data the display overhead in Emacs can be significantly
reduced and system throughput can be significantly improved.  Short
bursts of data will be buffered with a very small latency, however,
thus maintaining an interactive interface with a short response time.

The functional specification of abcat is simple: standard input is
copied to standard output.  The timing specification is a bit complex,
however.  Although abcat buffers the input stream, it also maintains
the idea of a guaranteed service time that changes depending on how
the input data arrives.  The buffered data is flushed to the output
stream when either the buffer is full or the service time has expired.
Assuming that no input data has arrived for a while, then as soon as a
chunk of data arrives it is output immediately.  After that,
subsequent data is held for a short while before being flushed out. If
data continues to arrive quickly, the arriving data is held for longer
and longer periods before being flushed, up to a maximum delay time.
If the input stream becomes quiet for a while then the buffer is
flushed and the next chunk of data is flushed immediately after its
arrival etc. 

The idea is to provide guaranteed quick service for small bursts of
data, to begin buffering when lots of data begins arriving very
quickly, and to always flush the buffer when data appears to have
stopped arriving.  If the parameters are set right for your system,
then for short interactions you should not even be aware that output
is being sent through abcat.  Once large quantities of data begin
streaming through, however, you may begin to see the effects of the
buffering delay.  At this point the data is probably streaming out
much faster than you can read it, however.  If for example you're only
looking for the Q.E.D at the end of an ACL2 run then this shouldn't be
a problem, but the amount of time that Emacs spends servicing the
incoming input stream should be significantly reduced.

Note that abcat knows nothing about files, and simply copies standard input to
standard output.  Therefore it must be used in a pipe, as in

% acl2 | abcat

Also note that abcat ignores the signal SIGINT (^c).  This allows the program
feeding abcat (e.g., ACL2) to process the signal.

To use abcat:

abcat [-s buffer_size] [-d max_delay] [-l service_levels] [-t hold_time] [-v]

The following parameters are user selectable:

-s buffer_size

  The size of the buffer in bytes, defaulting to 8192 bytes.

-d max_delay

  The maximum waiting time (in seconds) from when an input arrives to when it
  is flushed from the output buffer.  The max_delay parameter is a floating
  point number, and the default is 1.0 seconds.

-l service_levels

  The interval from 0 to max_delay seconds is divided into this many segments.
  If input data arrives continuously, where continuously means more
  frequently than the hold_time (see below), then the amount of time
  before the buffer is flushed increases in this many steps, until the
  maximum delay is reached.  This parameter is an integer > 0, and
  defaults to 4.

-t hold_time

  This is the maximum amount of time that abcat will hold a partially
  filled buffer without subsequent arrival of input.  In other words,
  if no input arrives for this amount of time then the buffer will be
  flushed.  This means that even if the max_delay is set to a
  relatively long time, once input appears to have stopped arriving
  then the buffer will be flushed anyway.  This parameter is a
  floating point number such that max_delay >= hold_time > 0.0, and
  the default is 0.25 seconds. 

-v

  If this option is selected then abcat will write some statistics
  about the run to standard error when the program terminates.  This
  information may help you set reasonable parameters for your system.

Note that the default parameter values were chosen heuristically and
empirically based on a small number of examples running under Linux.  I found
that buffers larger than the 8K default began to cause slightly decreased
performance on selected examples.  A max_delay of 1 second is sufficient to
fill the buffer pretty well and still give a warm fuzzy feeling that the run is
making progress.  The hold_time of .25 seconds is probably just over the limit
of a time interval that a human can perceive, so there is no feeling of lost
interactivity for short bursts.  The choice of 4 service levels gives fast
response for short bursts, but quickly accelerates buffering for increased
efficiency for longer bursts. 

  */
/* 
To get code for AIX uncomment the following:

#define AIX
*/

/*
Implementation Notes

  It seems that under AIX, select() will return a status indicating
  that there is input available on stdin, even though a subsequent
  read() will give the EAGAIN (no input available) error!  That's why in
  the code we always check for this error after a read(), and simply
  treat it as a 0-length read.

  */

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/times.h>
#include <sys/types.h>

#ifdef AIX
#include <sys/select.h>
#include <time.h>
#endif

/*  Default values of user-selectable parameters. */

#define BUFFER_SIZE 8192
#define MAX_DELAY 1.0
#define SERVICE_LEVELS 4
#define HOLD_TIME 0.25

#define STDIN 0
#define STDOUT 1

#define min(x, y) ((x) < (y) ? (x) : (y))

int buffer_size;		/* Buffer size. */
double max_delay;		/* Max delay in seconds between flushes. */
int service_levels;		/* Steps from 0 delay to max_delay. */
double hold_time;		/* Maximum time to hold buffer between
				   inputs */

char *ibuf;			/* Input buffer. */
int *delay_target;		/* Array of target time values. */
int hold_time_in_clocks;	/* Hold time in system clock units. */
int hold_sec, hold_usec;	/* Hold time values for select timer. */
int verbose;			/* Reporting flag. */

/* For Reporting */

int *nchpsl;			/* nchars per service level */
int *nevpsl;			/* nevents per service level */
int *nflpsl;			/* nflushes per service level */
int *nfupsl;			/* nfulls per service level */
int *nentsl;			/* netries to each service level */

/*  Initialization */

void initialize (argc, argv) 
     int argc;
     char *argv[];
{
  int c;
  int max_delay_in_clocks, interval_in_clocks;
  int i;
  double dhold_sec, dhold_usec;

  buffer_size = BUFFER_SIZE;
  max_delay = MAX_DELAY;
  service_levels = SERVICE_LEVELS;
  hold_time = HOLD_TIME;
  verbose = 0;

  while( ( c = getopt( argc, argv, "s:d:l:t:v") ) != EOF ) {
      switch ( c ) {
      case 's' : buffer_size = atoi( optarg); break;
      case 'd' : max_delay = atof( optarg); break;
      case 'l' : service_levels = atoi( optarg); break;
      case 't' : hold_time = atof( optarg); break;
      case 'v' : verbose = 1; break;
      default  : fprintf( stderr, "\
abcat: Usage: abcat [-s buffer_size] [-d max_delay] [-l service_levels]
abcat:              [-t hold_time] [-v]
abcat: Defaults: buffer_size = %d, %d service_levels, 
abcat:           max_delay = %.2f seconds, hold_time = %.2f seconds.\n",
			  BUFFER_SIZE, SERVICE_LEVELS, MAX_DELAY, HOLD_TIME);
      exit(1);
      }
    }

  if ( ( buffer_size <= 0 ) || ( max_delay <= 0.0 ) || 
       ( service_levels <= 0 ) || ( hold_time <= 0.0 ) ||
       ( max_delay < hold_time ) ) {
    fprintf( stderr, "abcat: Illegal parameter values.\n");
    exit(2);
  }

  ibuf = (char *)malloc( buffer_size * sizeof( char ) );
  delay_target = (int *)malloc( service_levels * sizeof( int ) );
  
  max_delay_in_clocks = (int)(max_delay * CLOCKS_PER_SEC);
  interval_in_clocks = max_delay_in_clocks / service_levels;
  hold_time_in_clocks = (int)(hold_time * CLOCKS_PER_SEC);

  if ( ( interval_in_clocks < 1 ) || ( hold_time_in_clocks < 1 ) ) {
    fprintf( stderr,"abcat: Timing specification below system granularity.\n");
    exit(3);
  }

  dhold_usec = modf( hold_time, &dhold_sec);
  hold_sec = (int) dhold_sec;
  hold_usec = min( (int)( dhold_usec * 1000000 ), 999999);

  for ( i = 0; i < service_levels; i++) {
    delay_target[i] = (i + 1) * interval_in_clocks;
  }
    
  /* For reporting */

  nchpsl = (int *)malloc ( service_levels * sizeof( int ) );
  nevpsl = (int *)malloc ( service_levels * sizeof( int ) );
  nflpsl = (int *)malloc ( service_levels * sizeof( int ) );
  nfupsl = (int *)malloc ( service_levels * sizeof( int ) );
  nentsl = (int *)malloc ( service_levels * sizeof( int ) );

  for (i = 0; i < service_levels; i++) {
    nchpsl[i] = 0;
    nevpsl[i] = 0;
    nflpsl[i] = 0;
    nfupsl[i] = 0;
    nentsl[i] = 0;
  }
  nentsl[0] = 1;		/* We begin in service_level 0. */
}

/* MAIN */

main (argc, argv) 
     int argc;
     char *argv[];
{
  char *iptr;			/* Current location in input buffer. */

#ifdef AIX
  int rdfs;
#else
  fd_set rdfs;			/* The Read File Set for select. */
#endif

  struct timeval tv, *tv_arg;   /* Stores select timeout delay. */
  int select_val;		/* Return value from select. */

  struct tms t;			/* Dummy structure for calling times */

  int nread, nwritten, nleft;	/* Characters read and written, space left in
				   buffer. */ 
  int service_level;		/* Selects timeout delay for writing partially
				   full buffers. */
  int eof;

  /*  Timers are in clock units returned by times. */

  int last_arrival, last_flush, now, service_wait;

  double integral_delay, fractional_delay;

  /* For reporting */

  int nchars = 0, nevents = 0, nflushes = 0, nfulls = 0;
  int last_service_level = 0; 
  int i;

  /*  BEGIN  ... the world's most complicated `cat' program ... */

  initialize( argc, argv);

  signal(SIGINT, SIG_IGN);	/* Ignore ^c signals. */

  fcntl( STDIN, F_SETFL, O_NONBLOCK); /* Set stdin for non-blocking reads. */

  service_level = 0;
  iptr = ibuf;
  nleft = buffer_size;
  last_flush = 0;
  last_arrival = 0;
  eof = 0;

#ifndef AIX
  FD_ZERO( &rdfs );		/* Setup select call. */
#endif
  tv_arg = NULL;		/* Tells select to sleep til input or EOF */

  do {

#ifdef AIX
    rdfs = 1;
#else
    FD_SET( STDIN, &rdfs);
#endif

    select_val = select( 1, &rdfs, NULL, NULL, tv_arg);

    /* If select_val == 0, then we were awakned by a timeout.  Otherwise,
       there is data to be read (or EOF or error). */

    if ( select_val < 0 ) {
      perror( "abcat - select");
      return 1;
    }

    now = times(&t);

    if ( select_val ) {
      
      nread = read( STDIN, iptr, nleft);
  
      switch ( nread ) {
      case 0 : eof = 1; break;
      case -1 :			
	if ( errno == EAGAIN ) { /* See Implementation Notes above. */
	  nread = 0;
	} else {
	  perror( "abcat - read"); return 1; break;
	}
      }

      iptr += nread;
      nleft -= nread;

      if ( verbose ) {
	if ( ! eof ) {
	  nevents++;
	  nevpsl[service_level]++;
	}
	nchars += nread;
	nchpsl[service_level] += nread;
      }
    }

    /*  The buffer is flushed if: 1) select timed out, 2) the buffer
	is full, 3) EOF, or 4) we observe that the delay target for this
	service level has been exceeded. */

    service_wait = ( delay_target[service_level] + last_flush ) - now;

    if ( ( ! select_val) || ( nleft == 0 ) || eof || ( service_wait <= 0 ) ) {

      if ( verbose ) {
	nflushes++;
	nflpsl[service_level]++;
	nfulls += ( nleft == 0 );
	nfupsl[service_level] += ( nleft == 0 );
      }

	nwritten = write( STDOUT, ibuf, iptr - ibuf); /* Flush buffer. */

	if ( nwritten != ( iptr - ibuf ) ) {
	  fprintf( stderr, "\nabcat: Read %d but wrote %d.\n", nread,
		   nwritten);
	  return 1;
	}
	iptr = ibuf;
	nleft = buffer_size;
	tv_arg = NULL;		
	last_flush = now;

	/*  The idea is to increase the `service level' (delay time) each time
	    we flush a buffer, up to a maximum amount.  However, if we
	    flushed the buffer after a hold timeout we reset the
	    service_level back to 0. */

	if ( verbose ) last_service_level = service_level;

	if ( (now - last_arrival) >= hold_time_in_clocks ) {
	    service_level = 0;
	  } else {
	    service_level = min( service_level + 1, service_levels - 1 );
	  }

	if ( verbose && ( service_level != last_service_level ))
	  nentsl[service_level]++;
    
    } else {

      /*  There was data and the buffer is not full.  We need to reschedule a
	  wakeup to flush the buffer.  We'll wake up either at the
	  time required by the service level or the hold time,
	  whichever is less.  We'll also wake up should any more data arrive in
	  the meantime. */

      tv_arg = &tv;
      
      if ( hold_time_in_clocks <= service_wait ) {

	tv.tv_sec = hold_sec;
	tv.tv_usec = hold_usec;

      } else {

	fractional_delay = modf( ((double)service_wait)/CLOCKS_PER_SEC, 
				 &integral_delay);
	tv.tv_sec = (int) integral_delay;
	tv.tv_usec = min( (int)(fractional_delay * 1000000), 999999);
      }
    }

    if ( select_val ) last_arrival = now;

  } while ( ! eof );

  if ( verbose ) {
	fprintf( stderr,"
abcat: Buffer size %d, %d service levels, max delay %.2f seconds,
abcat: hold time %.2f seconds.", buffer_size, service_levels, max_delay,
		 hold_time); 
	fprintf( stderr,"\nabcat: Copied %d chars. in %d events \
(avg. %.2f chars./event).", nchars, nevents, (double)nchars/nevents);
	fprintf( stderr,"\nabcat: Flushed buffer %d times (avg. %d chars./\
flush).", nflushes, nchars/nflushes);
	fprintf( stderr,"\nabcat: Buffer was full %d times.", nfulls);
	fprintf( stderr,"\nabcat: Level, entries,  chars,   events,    avg., \
 flushes, chars/flush, fulls.\n");
	for (i = 0; i < service_levels; i++) {
	  fprintf( stderr, "abcat: %2d %8d %10d %7d %10.2f %8d %10.2f %9d\n",
		   i, nentsl[i],
		   nchpsl[i], nevpsl[i],
		   nevpsl[i] == 0 ? 0 : (double)nchpsl[i]/nevpsl[i],
		   nflpsl[i],
		   nflpsl[i] == 0 ? 0 : (double)nchpsl[i]/nflpsl[i],
		   nfupsl[i]);
	}
  }
}

  

