;;;  pseudo-process.el -- Efficient process output via files

;;  This is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
    
;;  This code is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
    
;;  You should have received a copy of the GNU General Public License
;;  along with GNU Emacs; if not, write to the Free Software
;;  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;  Bishop Brock
;;  IBM Austin Research Laboratory
;;  11501 Burnet Rd. MS 9460
;;  Austin, TX 78758
;;  brock@austin.ibm.com
;;
;;  The process output handling code is a munged version of code stolen from
;;  the GNU Emacs version 19.34.1 comint package authored as follows:
;;
;;  Author: Olin Shivers <shivers@cs.cmu.edu>
;;  Adapted-by: Simon Marshall <simon@gnu.ai.mit.edu>
;;

;;  This collection of Emacs functions allows growing files to be treated as
;;  `pseudo-processes' which display their output in a buffer.  The idea is to
;;  reduce the overhead in Emacs when normal process buffers are used to
;;  display output from processes generating a lot of output (like ACL2).
;;  Using pseudo-process output can result in drastic savings of overhead for
;;  processes that produce large amounts of output.  This savings far surpasses
;;  the relativey small amount of overhead involved in writing the output to
;;  an intermediate file. In fact, if the buffer is not visible in any window,
;;  output display is delayed until the buffer becomes visible again. Thus one
;;  can run big, output intensive jobs with practically zero display overhead
;;  by simply hiding the buffer until the job is finished.
;;
;;  To use this facility, one directs the real output of the process to a file,
;;  then invokes M-x start-pseudo-process and gives the process output
;;  filename.  As the file grows, the output will be displayed in the buffer
;;  current when start-pseudo-process was invoked.  The pseudo-process is
;;  cancelled when the buffer is killed, or can be explicity killed with M-x
;;  cancel-pseudo-process. 
;;
;;  If you are in a comint-mode shell, you can also do M-x pseudo-command which
;;  will prompt for a string to pass to the shell as a command.  This command
;;  will be appended with ">& <filename>", where <filename> is a unique
;;  filename generated by pseudo-command.
;;
;;  If the buffer was a comint-mode buffer, the pseudo-process output handler
;;  will try to maintain all the markers that comint mode uses, so that the
;;  comint-mode commands will work as usual.  Note however that by using
;;  pseudo-process output you lose the ability to filter the output (unless you
;;  modify the function pseudo-process-handler-function).
;;
;;  You can also use this facility to follow along with any growing file. e.g.,
;;  a log file.  The command M-x pseudo-follow will prompt you for a filename,
;;  then begin display of that file in a buffer named *Follow <filename>*.  If
;;  a buffer of this name already exists and appears to be following the same
;;  file (presumably an old version), the command will confirm that you want to
;;  re-use the buffer.  Otherwise, a buffer *Follow <filename>*<2> etc will be
;;  created.  In all cases the file must already exist before you can begin
;;  following it.
;;
;;  Although the code works well enough for my purposes, there are
;;  undoubtedly numerous ways to confuse this code.  For example, if you change
;;  the major mode of the buffer all of the local variables will probably get
;;  killed, with unpredicatable results to the pseudo-process.
;;
;;  Here are some things you can customize:

(defconst pseudo-process-minimum-interval 0.25
  "*Minimum interval in seconds between display updates from
pseudo-processes.  Can be an integer or floating point number > 0.  If the
pseudo-process file grows continuously, subsequent display updates
are delayed in increments of this minimum interval up to the maximum
interval.") 

(defconst pseudo-process-maximum-interval 1.0
  "*Maximum interval in seconds between display updates from
pseudo-processes.  Can be an integer or floating point number > 0.  Must also
be >= pseudo-process-minimum-interval.  This is the maximum amount of time
(give or take) that pseudo-process output will be delayed when the
pseudo-process file is growing rapidly.") 

(defvar pseudo-process-directory "/tmp"
  "*If you like you can set this variable to a string containing a default
directory for pseudo-process files.  The default setting is /tmp.")

(defun pseudo-process-filename (&optional string)
  "Use the current-time to create a filename unique to the microsecond of it's
creation, containing an optional string."
  (let* ((time (current-time))
	 (file (format "%s%spp_%s_%d_%d_%d~"
		       (if pseudo-process-directory
			   pseudo-process-directory 
			 "")
		       (if pseudo-process-directory "/" "")
		       (if (stringp string) string "")
		       (car time)
		       (car (cdr time))
		       (car (cdr (cdr time))))))
    file))
  
;;  Start Code.

;;  The following variables will all be local to a buffer while running a
;;  pseudo-process.

(defvar pseudo-process-timer nil
  "Timer object for this pseudo-process.")
(make-variable-buffer-local 'pseudo-process-timer)

(defvar pseudo-process-file nil
  "File for this pseudo-process.")
(make-variable-buffer-local 'pseudo-process-file)

(defvar pseudo-process-position nil
  "Current position in the file for pseudo-process-output.")
(make-variable-buffer-local 'pseudo-process-position)

(defvar pseudo-process-delete-on-exit nil
  "If non-NIL then the pseudo-process file will be deleted when the
pseudo-p rocess is cancelled.")
(make-variable-buffer-local 'pseudo-process-delete-on-exit)

(defvar pseudo-process-suspended nil
  "Set when pseudo-process output is inhibited due to invisibility of the
buffer.")
(make-variable-buffer-local ' pseudo-process-suspended)

(defvar pseudo-process-max-service-level 0
  "Number of steps in the delay increment from the minimum to the maximum
delay.")
(make-variable-buffer-local 'pseudo-process-max-service-level)

(defvar pseudo-process-service-level 0
  "Current service level for this buffer.")
(make-variable-buffer-local ' pseudo-process-service-level)

(defvar pseudo-process-countdown 0
  "Counts down timer events until 0, at which point new data is displayed.")
(make-variable-buffer-local 'pseudo-process-countdown)

(defvar pseudo-process-last-size 0
  "The filesize the last time the pseudo-process handler was invoked.")
(make-variable-buffer-local 'pseudo-process-last-size)

(defvar pseudo-process-mark nil
  "Pseudo process-mark for non comint-mode buffers.")
(make-variable-buffer-local 'pseudo-process-mark)

(defvar pseudo-process-last-output-start nil
  "For non-comint-mode buffers, a marker used to decide whether to continue
pushing point along as the file fills the buffer.")
(make-variable-buffer-local 'pseudo-process-last-output-start)

(defun start-pseudo-process (file &optional min-interval max-interval
				  createp deletep)

  "Start pseudo-process output from the given file into the current buffer, and
cancel any other pseudo-process running in this buffer.

The min-interval is the number of seconds (integer or floating point) to wait
between display updates.  If the interval is unspecified or NIL, the default
value `pseudo-process-minimum-interval' will be used.  

If the file grows rapidly, subsequent display updates are delayed in steps of
the minimum interval up to the max-interval.  If the max-interval is
unspecified or NIL, AND the min-interval was specified, then the max-interval
defaults to the min-interval.  Otherwise, the max-interval defaults to the
value of pseudo-process-maximum-interval.

If createp is non-nil, the file will be re(created) and initialized to 0
length. 

If deletep is non-NIL, the file will be deleted when the pseduo-process is
cancelled. 

A buffer can have at most one pseudo-process runnning at a time.  The
pseduo-process can be cancelled with M-x cancel-pseudo-process.  The
pseudo-process is also cancelled when the buffer is killed.

If the buffer is a comint-mode buffer then the pseudo-process update routine
will try to maintain all the marks that the standard comint filter
maintains. Otherwise, the pseudo `process mark' for non-process buffers will be
initialized at the current value of point in the buffer."

  (interactive "FFile : \n")

  (cond
   ((not (stringp file)) 
    (error "start-pseudo-process: Filename not a string."))
   ((and min-interval (or (not (numberp min-interval)) (<= min-interval 0)))
    (error "start-pseudo-process: Illegal minimum interval."))
   ((and (not min-interval) (or (not (numberp pseudo-process-minimum-interval))
				(<= pseudo-process-minimum-interval 0)))
    (error "start-pseudo-process: Illegal default minimum interval."))
   ((and max-interval (or (not (numberp max-interval)) (<= max-interval 0)))
    (error "start-pseudo-process: Illegal maximum interval."))
   ((and (not min-interval) (not max-interval)
	 (or (not (numberp pseudo-process-maximum-interval))
	     (< pseudo-process-maximum-interval 0)))
    (error "start-pseudo-process: Illegal default maximum interval."))

   (t
    (let ((min-interval (if min-interval
			    min-interval
			  pseudo-process-minimum-interval))
	  (max-interval (if max-interval
			    max-interval
			  (if min-interval
			      pseudo-process-minimum-interval
			    pseudo-process-maximum-interval))))
      (if (< max-interval min-interval)
	  (error "start-pseudo-process:  max-interval < min-interval"))

      (cancel-pseudo-process)

      (if createp (write-region "" "" file))

      (if (not (file-attributes file))
	  (error "start-pseudo-process: File %s can't be located." file))
    
      (setq pseudo-process-file file)
      (setq pseudo-process-position 0)
      (setq pseudo-process-delete-on-exit deletep)
      (setq pseudo-process-suspended nil)
      (setq pseudo-process-max-service-level 
	    (floor (/ (- max-interval min-interval) min-interval)))
      (setq pseudo-process-service-level 0)
      (setq pseudo-process-countdown 0)

      ;;  We assume that if this is really a process buffer, then the user
      ;;  has started the real process before starting the pseudo-process.

      (let ((process (get-buffer-process (current-buffer))))
	(if process
	    (progn
	      (setq pseudo-process-mark (process-mark process))
	      (if (and (boundp 'comint-last-output-start)
		       comint-last-output-start)
		  (setq pseudo-process-last-output-start
			comint-last-output-start)
		(progn
		  (setq pseudo-process-last-output-start (make-marker))
		  (set-marker pseudo-process-last-output-start (point-min)))))
	  (progn
	    (setq pseudo-process-last-output-start (make-marker))
	    (setq pseudo-process-mark (make-marker))
	    (set-marker pseudo-process-last-output-start (point-min))
	    (set-marker pseudo-process-mark (point))))

	(let ((timer
	       (run-at-time 0 min-interval 'pseudo-process-handler
			    (current-buffer))))
	  (setq pseudo-process-timer timer)))))))

(add-hook 'kill-buffer-hook 'cancel-pseudo-process)

(defun cancel-pseudo-process ()
  "Cancel the pseudo-process for the current buffer."
  (interactive)
  (condition-case nil
    (if (and (boundp 'pseudo-process-timer)
	     pseudo-process-timer)
	(progn (cancel-timer pseudo-process-timer)
	       (if pseudo-process-delete-on-exit
		   (delete-file pseudo-process-file))))
    (error nil)))

(add-hook 'kill-emacs-hook 'cancel-all-pseudo-processes)

(defun cancel-all-pseudo-processes ()
  "Cancel all pseudo-processes from a scan of the buffer-list."
  (interactive)
  (save-excursion
    (mapcar (function (lambda (buffer)
			(set-buffer buffer)
			(cancel-pseudo-process)))
	    (buffer-list))))

;; If the pseudo-process handler crashes for any reason it causes all kinds of
;; problems, so we need to set up a highly protected call.  We only do a
;; save-excursion if we're not updating the current buffer, since the idea is
;; to change point and mark.

(defun pseudo-process-handler (buffer)
  (cond
   ((eq buffer (current-buffer))
    (condition-case nil
      (pseudo-process-handler-function buffer)
      (error (cancel-pseudo-process))))
   (t (condition-case nil
	(save-excursion
	  (set-buffer buffer)
	  (pseudo-process-handler-function buffer))
	(error nil)))))

(defun pseudo-process-handler-function (buffer)
  "Update a buffer from a growing file. Similar to comint-output-filter.
Buffer has been set to this buffer before the call.  We're protected outside
from all errors."

  (cond

   ;;  If the buffer is not even visible we quit right away.

   ((not (get-buffer-window buffer 'visible))
    (setq pseudo-process-suspended t)
    (setq pseudo-process-service-level 0))

   (t (let* ((file-info (file-attributes pseudo-process-file))
	     (size (nth 7 file-info)))
	(cond
	 ((or (not file-info) (< size pseudo-process-position))
	  (cancel-pseudo-process)
	  (error "Pseudo-process: File %s on display in buffer %s does not exist or appears to have shrunk. Aborting pseudo-process for this buffer."
		 pseudo-process-file (buffer-name buffer)))

	 ;;  If the file size is unchanged, there's nothing to do.

	 ((= size pseudo-process-position) 
	  (setq pseudo-process-service-level 0)
	  (setq pseudo-process-suspended nil))

	 ;;  If we're in a multiple-interval delay, and data has continued to
	 ;;  arrive, we continue waiting.

	 ((and (< 0 pseudo-process-service-level)
	       (< 0 pseudo-process-countdown)
	       (not (= pseudo-process-last-size size)))
	  (setq pseudo-process-countdown (1- pseudo-process-countdown))
	  (setq pseudo-process-last-size size))

	 (t

	  ;;  This code was stolen and modified from comint-output-filter.
	  ;;  There, output is always into a process buffer.  Here, we may be
	  ;;  writing into a normal buffer. 

	  (let ((opoint (point))
		(obeg (point-min))
		(oend (point-max))
		(buffer-read-only nil)
		(ostart nil)
		(process (get-buffer-process buffer))
		(nchars (- size pseudo-process-position)))

	    (widen)
	    (goto-char pseudo-process-mark)

	    ;;  Output starts at the process mark, and we adjust a narrowed
	    ;;  region for the new stuff to be added.

	    (setq ostart (point))
	    (if (<= (point) opoint)
		(setq opoint (+ opoint nchars)))
	    ;; Insert after old_begv, but before old_zv.
	    (if (< (point) obeg)
		(setq obeg (+ obeg nchars)))
	    (if (<= (point) oend)
		(setq oend (+ oend nchars)))

	    ;;  Here the new bytes are loaded into the buffer. Note that we
	    ;;  want to mimic `insert-before-markers'.  To do this we first
	    ;;  insert a single character to move all of the markers, insert
	    ;;  the file, then delete the char.

	    (if pseudo-process-suspended
		(princ 
		 (format "Please stand by while %d new characters are loaded." 
			 nchars)
		 t))

	    (insert-before-markers ?0)
	    (goto-char (- (point) 1))
	    (insert-file-contents-literally 
	     pseudo-process-file nil pseudo-process-position size)
	    (goto-char (+ (point) nchars))
	    (delete-char 1)

	    (if pseudo-process-suspended
		(progn
		  (princ " Done" t)
		  (setq pseudo-process-suspended nil)))

	    (setq pseudo-process-position size)

	    ;;  If data arrived in the last interval we bump up the delay
	    ;;  time. Otherwise, it's reset to 0.

	    (setq pseudo-process-service-level 
		  (if (not (= size pseudo-process-last-size))
		      (min (1+ pseudo-process-service-level)
			   pseudo-process-max-service-level)
		    0))

	    (setq pseudo-process-countdown pseudo-process-service-level)
	    (setq pseudo-process-last-size size)

	    ;; Don't insert initial prompt outside the top of the window.

	    (if (and (eq (window-buffer (selected-window)) buffer)
		     (= (window-start (selected-window)) (point)))
		(set-window-start (selected-window) (- (point) nchars)))

	    ;;  We don't maintain a `last-input-end' for non-comint buffers.
	    
	    (if (and (boundp 'comint-last-input-end)
		     comint-last-input-end
		     (marker-buffer comint-last-input-end)
		     (= (point) comint-last-input-end))
		(set-marker comint-last-input-end 
			    (- comint-last-input-end nchars)))

	    (set-marker pseudo-process-last-output-start ostart)

	    (force-mode-line-update)

	    (narrow-to-region obeg oend)
	    (goto-char opoint)

	    ;;  This is a munged version of the standard hook
	    ;;  comint-postoutput-scroll-to-bottom 

	    (let* ((selected (selected-window))
		   (scroll (if (boundp 'comint-scroll-to-bottom-on-output)
			       comint-scroll-to-bottom-on-output
			     nil))
		   (process-mark-visible-p 
		    (and (<= (point-min) pseudo-process-mark)
			 (<= pseudo-process-mark (point-max)))))
	      (walk-windows
	       (function 
		(lambda (window)
		  (if (eq (window-buffer window) buffer)
		      (progn
			(select-window window)
			(if (and 
			     (< (point) pseudo-process-mark)
			     process-mark-visible-p
			     (or (eq scroll t) (eq scroll 'all)
				 ;; Maybe user wants point to jump to
				 ;; end.
				 (and (eq scroll 'this) 
				      (eq selected window)) 
				 (and (eq scroll 'others) 
				      (not (eq selected window)))
				 ;; If point was at the end, keep it at
				 ;; end.
				 (>= (point)
				     pseudo-process-last-output-start))) 
			    (goto-char pseudo-process-mark))
			;; Optionally scroll so that the text
			;; ends at the bottom of the window.
			(if (and (boundp 'comint-scroll-show-maximum-output)
				 comint-scroll-show-maximum-output
				 (>= (point) pseudo-process-mark)
				 process-mark-visible-p)
			    (save-excursion
			      (goto-char (point-max))
			      (recenter -1)))
			(select-window selected)))))
	       nil t)))))))))

(defun pseudo-command (command)
  "Run some command (presumably from a comint-mode shell buffer), using
pseudo-process output.  An intermediate file will be (re)created for the
command output, then deleted when the buffer is killed or the pseudo-process is
cancelled."

  (interactive "sCommand : \n")

  (let ((process (get-buffer-process (current-buffer))))
    (if (not process)
	(error "Not a process buffer.")
      (let* ((file (pseudo-process-filename (buffer-name (current-buffer)))))
	(start-pseudo-process file nil nil t t)
	(insert (format "%s >& '%s'\n" command file))
	(comint-send-input)))))

(defun pseudo-follow (filename)
  "Follow the growth of the file in a pseudo-process buffer named 
*Follow <filename>*"
  (interactive "fFollow file : ")
  (let ((buffer-name (concat "*Follow " filename "*"))
	buffer)
    (cond
     ((get-buffer buffer-name)
      (save-excursion
	(set-buffer buffer-name)
	(cond
	 ((and (boundp 'pseudo-process-file)
	       pseudo-process-file
	       (equal pseudo-process-file filename))
	  (switch-to-buffer buffer-name)
	  (cond
	   ((yes-or-no-p "Kill this pseudo-follow of the same file?")
	    (set-buffer-modified-p nil)
	    (kill-buffer buffer-name))))))))
    (setq buffer (generate-new-buffer buffer-name))
    (switch-to-buffer buffer)
    (start-pseudo-process filename)))
    
;;  The M-x pseudo-acl2 command runs ACL2 as a pseudo process.  To get this to
;;  work you need to define a general-purpose command file that accepts a
;;  command name and an output file name. Uncomment the following 2 lines and
;;  put them in a file called `pseudo-command' somewhere along your execution
;;  path.  Be sure to set the execute permission on the file after doing this.

;#!/bin/sh
;$1 1>> $2 2>> $2

(defun pseudo-acl2 (&optional command buffer-name)
  "Run acl2 in a buffer named *acl2*. The acl2 will run as a pseudo
process. With argument, allows you to select an acl2 to run."
  (interactive (list (if current-prefix-arg
			 (read-string "ACL2 Command :" "acl2")
		       "acl2")))
  (require 'inf-lisp)
  (let (real-buffer-name)
    (setq command (if (or (not command) (equal command "")) "acl2" command))
    (setq buffer-name (if buffer-name buffer-name command))
    (setq real-buffer-name (concat "*" buffer-name "*"))
    (while (get-buffer real-buffer-name)
      (let ((prompt-string
	     (format "Buffer %s exists, choose another name (without `*') : "
		     real-buffer-name)))
	(beep)
	(setq buffer-name (read-string prompt-string))
	(setq real-buffer-name (concat "*" buffer-name "*"))))

    ;;  We create the file explicitly here because of the unknown time delay
    ;;  between make-comint and the actual firing up of the process.

    (let* ((file (pseudo-process-filename real-buffer-name))
	   (ignore (write-region "" "" file))
	   (buffer (make-comint buffer-name "pseudo-command" nil command
				file)))
      (set-buffer buffer)
      (inferior-lisp-mode)
      (setq inferior-lisp-buffer buffer-name)
      (let ((pop-up-windows nil)) (pop-to-buffer buffer))
      (start-pseudo-process file nil nil nil t))))
